#!/usr/bin/env python3
#
#


import influxdb_client
import matplotlib.pyplot as plt
import numpy as np
import os
import smtplib
import sys

from datetime import datetime, timedelta
from email.message import EmailMessage
from email.utils import make_msgid
from influxdb_client.client.write_api import SYNCHRONOUS


def run_query(q, query_api):
    ''' Run a query and return the result
    '''
    print(q)
    tables = query_api.query(q)
    return tables
    

def gen_pie_chart(data, labels, name):
    y = np.array(data)
    
    plt.pie(y, labels = labels, shadow = True, autopct='%.1f%%')
    plt.tight_layout()
    plt.savefig(f'{name}.png')
    plt.figure().clear()
    plt.close()
    plt.cla()
    plt.clf()
    

BUCKET=os.getenv("BUCKET")
LOW_GRAN_BUCKET=os.getenv("LOW_GRAN_BUCKET")
ORG=os.getenv("ORG")
MAIL = {
    "to": os.getenv("MAIL_TO"),
    "from": os.getenv("MAIL_FROM"),
    "server": os.getenv("MAIL_SRV"),
    "port": os.getenv("MAIL_PORT"),
    "user": os.getenv("MAIL_USER"),
    "password" : os.getenv("MAIL_PASS")
    }
    
# Setup
client = influxdb_client.InfluxDBClient(url="http://telegrafexternal.bentasker.co.uk", token="", org="", timeout=120_000)
query_api = client.query_api()

# TODO: control this via commandline argument
DAYS=int(sys.argv[1])

AGG_DAYS="1d"

if DAYS > 30:
    # Use the low granularity bucket
    BUCKET = LOW_GRAN_BUCKET
    AGG_DAYS = "30d"

# generate the start to pass into `range()`
START=f"-{DAYS}d"

dt = datetime.now()
delta = timedelta(days = DAYS)
start_dt = dt - delta



data = {}

# Pageviews
data['pageviews'] = {"total": 0, "domains": {}}
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics" and r._field == "requests")
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> group()
  |> sum()
'''

data['pageviews']["total"] = run_query(q, query_api)[0].records[0].values['_value']

q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics" and r._field == "requests")
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> group(columns: ["domain"])
  |> sum()
'''
tables = run_query(q, query_api)
for table in tables:
    d = table.records[0].values["domain"]
    v = table.records[0].values["_value"]
    data['pageviews']['domains'][d] = v
    

q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "platform_PageViews")
  |> group(columns: ["domain","platform"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()  
  '''
  
tables = run_query(q, query_api)
data['pageviews']['platforms'] = {
    "domains" : {},
    "platforms" : {},
    }
for table in tables:
    d = table.records[0].values["domain"]
    p = table.records[0].values["platform"]
    v = table.records[0].values["_value"]
    
    if d not in data['pageviews']['platforms']["domains"]:
        data['pageviews']['platforms']["domains"][d] = {}
    
    if p not in data['pageviews']['platforms']['platforms']:
        data['pageviews']['platforms']['platforms'][p] = 0
    
    data['pageviews']['platforms']["domains"][d][p] = v
    data['pageviews']['platforms']['platforms'][p] += v
    

q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "tz_PageViews")
  |> group(columns: ["domain","timezone"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  '''
tables = run_query(q, query_api)  
data['pageviews']['timezones'] = {
    "domains" : {},
    "timezones" : {},
    }  
for table in tables:
    d = table.records[0].values["domain"]
    t = "tz_" + table.records[0].values["timezone"]
    v = table.records[0].values["_value"]
    
    if d not in data['pageviews']['timezones']["domains"]:
        data['pageviews']['timezones']["domains"][d] = {}
    
    if p not in data['pageviews']['timezones']['timezones']:
        data['pageviews']['timezones']['timezones'][t] = 0
    
    data['pageviews']['timezones']["domains"][d][t] = v
    data['pageviews']['timezones']['timezones'][t] += v


data['pageviews']['toppages'] = []
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 20)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['toppages'].append(r)


# Top bentasker.co.uk pages
data['pageviews']['topbentasker'] = []
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => r.domain == "www.bentasker.co.uk")
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['topbentasker'].append(r)

# Top snippets
data['pageviews']['snippets'] = []
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => r.domain == "snippets.bentasker.co.uk")
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['snippets'].append(r)



data['pageviews']['recipes'] = []
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => r.domain == "recipebook.bentasker.co.uk")
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['recipes'].append(r)


# Top photos
data['pageviews']['photosarchive'] = []
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => r.domain == "photosarchive.bentasker.co.uk")
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['photosarchive'].append(r)


# Top Onions
data['pageviews']['onions'] = []
q = '''
import "strings"

from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => strings.hasSuffix(v: r.domain, suffix: ".onion"))
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['onions'].append(r)
    

# Top eepsites
data['pageviews']['i2p'] = []
q = '''
import "strings"

from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => strings.hasSuffix(v: r.domain, suffix: ".i2p"))
  |> group(columns: ["domain", "page"])
  |> aggregateWindow(every: ''' + AGG_DAYS + ''', fn: sum)
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
for row in tables[0].records:
    r = [
        row.values['domain'],
        row.values['page'],
        row.values['_value']
        ]
    data['pageviews']['i2p'].append(r)
    

# Response times
data['response_times'] = {"mean" : {}, "min": {}, "max" : {}}
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics" and r._field == "response_time")
  |> filter(fn: (r) => r.action == "\\"PageView\\"")
  |> group(columns: ["domain"])
  |> mean()
'''

tables = run_query(q, query_api)
t = 0
c = 0
for table in tables:
    d = table.records[0].values["domain"]
    v = table.records[0].values["_value"]
    t += v
    c += 1
    data['response_times']['mean'][d] = v
    
data['response_times']['mean']["_mean"] = t / c    

q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics" and r._field == "min_response_time")
  |> filter(fn: (r) => r.action == "\\"PageView\\"")
  |> filter(fn: (r) => r._value > 0)
  |> group(columns: ["domain"])
  |> min()
'''
tables = run_query(q, query_api)
t = 0
c = 0
for table in tables:
    d = table.records[0].values["domain"]
    v = table.records[0].values["_value"]
    t += v
    c += 1
    data['response_times']['min'][d] = v
    
data['response_times']['min']["_mean"] = t / c


q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics" and r._field == "max_response_time")
  |> filter(fn: (r) => r.action == "\\"PageView\\"")
  |> filter(fn: (r) => r._value > 0)
  |> group(columns: ["domain"])
  |> max()
'''
tables = run_query(q, query_api)
t = 0
c = 0
for table in tables:
    d = table.records[0].values["domain"]
    v = table.records[0].values["_value"]
    t += v
    c += 1
    data['response_times']['max'][d] = v
    
data['response_times']['max']["_mean"] = t / c


# Referrers
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._field == "referrer_PageViews")
  |> group(columns: ["referringdomain"])
  |> sum()
  |> group()
  |> sort(columns: ["_value"], desc: true)
  |> limit(n: 20)
'''
tables = run_query(q, query_api)

data['referrers'] = []
for row in tables[0].records:
    r = [
        row.values['referringdomain'],
        row.values['_value']
        ]
    data['referrers'].append(r)


# Videos

data['videos'] = {"video_ready":{}}
q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._field == "response_time")
  |> filter(fn: (r) => r.action == "\\"ready\\"")
  |> group(columns: ["domain"])
  |> mean()
''' 
data['videos']["video_ready"]["mean"] = run_query(q, query_api)[0].records[0].values['_value']


q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._field == "min_response_time")
  |> filter(fn: (r) => r.action == "\\"ready\\"")
  |> filter(fn: (r) => r._value > 0)
  |> group(columns: ["domain"])
  |> min()
'''
data['videos']["video_ready"]["min"] = run_query(q, query_api)[0].records[0].values['_value']


q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._field == "max_response_time")
  |> filter(fn: (r) => r.action == "\\"ready\\"")
  |> filter(fn: (r) => r._value > 0)
  |> group(columns: ["domain"])
  |> max()
'''
data['videos']["video_ready"]["max"] = run_query(q, query_api)[0].records[0].values['_value']



q = '''
from(bucket: "''' + BUCKET + '''")
  |> range(start: ''' + START + ''')
  |> filter(fn: (r) => r._measurement == "pf_analytics")
  |> filter(fn: (r) => r._field == "requests")
  |> filter(fn: (r) => r.action == "\\"playing\\"")
  |> group(columns: ["page"])
  |> sum()
  |> group()
  |> sort(desc: true)
  |> limit(n: 10)  
'''
tables = run_query(q, query_api)
data['videos']['top'] = []
for row in tables[0].records:
    r = [
        row.values['page'],
        row.values['_value']
        ]
    data['videos']['top'].append(r)

print(data)



# Generate some graphs

domains = []
labels = []

domains_onion = []
labels_onion = []


for k in data['pageviews']['domains']:
    if k.endswith(".onion"):   
        domains_onion.append(data['pageviews']['domains'][k])
        
        # abbreviate onions to the first 6 chars        
        name = f'''{k[0:6]}...onion'''
        labels_onion.append(name)
    else:
        domains.append(data['pageviews']['domains'][k])
        # Abbreviate subdomains of bentasker.co.uk
        name = k.replace(".bentasker.co.uk",".bentasker")
        labels.append(name)        
        
        
gen_pie_chart(domains, labels, "domain_share")
gen_pie_chart(domains_onion, labels_onion, "domain_share_onion")

platforms = []
labels = []
for k in data['pageviews']['platforms']['platforms']:
    platforms.append(data['pageviews']['platforms']['platforms'][k])
    labels.append(k)    

gen_pie_chart(platforms, labels, "platform_share")


tzs = []
labels = []
for k in data['pageviews']['timezones']['timezones']:
    
    tz_n = str(round(int(k.replace("tz_","")) / 60, 0)).split(".")[0]
    tz = f"GMT {tz_n}h"

    # Because we're rounding, there may already be an entry for this tz
    # merge them if so
    #
    # index() returning a valueerror instead of False is dumb...
    try:
        idx = labels.index(tz)
        tzs[idx] += data['pageviews']['timezones']['timezones'][k]
    except:
        tzs.append(data['pageviews']['timezones']['timezones'][k])
        labels.append(tz)    

gen_pie_chart(tzs, labels, "timezones")




### Generate the report
css = '''
th, td {text-align: left}
table {max-width: 80%}
'''

tmplate = [ f'''
<style type="text/css">
{css}
</style>
           
           
<h1>PF Analytics {DAYS}d Report</h1>
Dates: <em>{start_dt.strftime("%Y-%m-%d")} - {dt.strftime("%Y-%m-%d")}</em><br />
<br />
<h3>Page Performance</h3>
<b>Total Page views</b>: {data['pageviews']["total"]}<br />
<table>
<tr><th></th><th>Page Views</th></tr>
''']

for rec in data['pageviews']['toppages']:
    pg = rec[1].replace('"','')
    tmplate.append(f'<tr><td>{rec[0]}{pg}</td><td>{rec[2]}</td></tr>')

domain_share_ccid = make_msgid()
onion_domain_share_ccid = make_msgid()

tmplate.append( f'''
</table>
<br />
<h3>Domain Traffic Share</h3>
<img src="cid:{domain_share_ccid[1:-1]}" /><br />
<b>Onions</b><br />
<img src="cid:{onion_domain_share_ccid[1:-1]}" /><br />

<br />
<h3>Response Times</h3>
<b>Average Response Time</b>: {round(data['response_times']['mean']["_mean"]/1000,2)}<br />

<table>
<tr><th></th><th>Mean Response Time</th></tr>
''')

for domain in data['response_times']['mean']:
    if domain == "_mean":
        continue
    t = round(data['response_times']['mean'][domain] / 1000,2)
    tmplate.append(f'''<tr><td>{domain}</td><td>{t}</td></tr>''')


platform_share_ccid = make_msgid()
timezone_share_ccid = make_msgid()

tmplate.append( f'''
</table>
<hr />
<h2>Traffic Sources</h2>
<h3>Platforms</h3><br />
<img src="cid:{platform_share_ccid[1:-1]}" /><br />
<table>
<tr><th></th><th>Page Views</th></tr>
''')

for k in data['pageviews']['platforms']['platforms']:
    tmplate.append(f'''<tr><td>{k}</td><td>{data['pageviews']['platforms']['platforms'][k]}</td></tr>''')

tmplate.append( f'''
</table><br />
<h3>Timezones</h3><br />
<img src="cid:{timezone_share_ccid[1:-1]}" /><br />
<table>
<tr><th></th><th>Page Views</th></tr>
''')

for k in data['pageviews']['timezones']['timezones']:
    tmplate.append(f'''<tr><td>{k}</td><td>{data['pageviews']['timezones']['timezones'][k]}</td></tr>''')

tmplate.append( f'''
</table><br />
<h3>Top 10 Referrers</h3>
<table>
<tr><th></th><th>Page Views</th></tr>
''')
for r in data['referrers']:
    tmplate.append(f'''<tr><td>{r[0]}</td><td>{r[1]}</td></tr>''')
    
tmplate.append( f'''
</table>
<br />
<hr />
<h3>Top Videos</h3>
<br />
<b>Min video ready time</b>: {round(data['videos']["video_ready"]["min"] / 1000,2)}<br />
<b>Average video ready time</b>: {round(data['videos']["video_ready"]["mean"] / 1000,2)}<br />
<b>Max video ready time</b>: {round(data['videos']["video_ready"]["max"] / 1000,2)}<br />
<table>
<tr><th></th><th>Playbacks</th></tr>
''')

for r in data['videos']['top']:
    tmplate.append(f'''<tr><td>{r[0].replace('"','')}</td><td>{r[1]}</td></tr>''')

tmplate.append( f'''
</table>
<br />
<h3>Top bentasker.co.uk content</h3>
<table>
<tr><th></th><th>Views</th></tr>
''')
for r in data['pageviews']['topbentasker']:
    tmplate.append(f'''<tr><td>{r[1].replace('"','')}</td><td>{r[2]}</td></tr>''')

tmplate.append( f'''
</table>
<br />
<h3>Top Snippets</h3>
<table>
<tr><th></th><th>Views</th></tr>
''')

for r in data['pageviews']['snippets']:
    tmplate.append(f'''<tr><td>{r[1].replace('"','')}</td><td>{r[2]}</td></tr>''')

tmplate.append( f'''
</table>
<br />
<h3>Top Recipes</h3>
<table>
<tr><th></th><th>Views</th></tr>
''')

for r in data['pageviews']['recipes']:
    tmplate.append(f'''<tr><td>{r[1].replace('"','')}</td><td>{r[2]}</td></tr>''')

tmplate.append( f'''
</table>
<br />
<h3>Top Photos</h3>
<table>
<tr><th></th><th>Views</th></tr>
''')

for r in data['pageviews']['photosarchive']:
    tmplate.append(f'''<tr><td>{r[1].replace('"','')}</td><td>{r[2]}</td></tr>''')

tmplate.append( f'''
</table>
<br />
<h3>Top Onion Content</h3>
<table>
<tr><th></th><th>Views</th></tr>
''')

for r in data['pageviews']['onions']:
    tmplate.append(f'''<tr><td>{r[0]}{r[1].replace('"','')}</td><td>{r[2]}</td></tr>''')
    
tmplate.append( f'''
</table>
<br />
<h3>Top eepsite Content</h3>
<table>
<tr><th></th><th>Views</th></tr>
''')


for r in data['pageviews']['i2p']:
    tmplate.append(f'''<tr><td>{r[0]}{r[1].replace('"','')}</td><td>{r[2]}</td></tr>''')
    
tmplate.append( f'''
</table>
<br />
''')


body = '\n'.join(tmplate)


msg = EmailMessage()
msg['to'] = MAIL['to']
msg['from'] = MAIL['from']
msg['subject'] = f'''Analytics {DAYS}d Summary: {start_dt.strftime("%Y-%m-%d")} - {dt.strftime("%Y-%m-%d")}'''

msg.set_content(body, 'html')

with open('domain_share.png','rb') as fp:
    msg.add_related(
        fp.read(), 'image', 'png', cid=domain_share_ccid)


with open('domain_share_onion.png','rb') as fp:
    msg.add_related(
        fp.read(), 'image', 'png', cid=onion_domain_share_ccid)
    
with open('platform_share.png','rb') as fp:
    msg.add_related(
        fp.read(), 'image', 'png', cid=platform_share_ccid)
    
with open('timezones.png','rb') as fp:
    msg.add_related(
        fp.read(), 'image', 'png', cid=timezone_share_ccid)

print(msg.as_string())

smtp_client = smtplib.SMTP_SSL(MAIL["server"],MAIL["port"])
smtp_client.login(MAIL["user"], MAIL["password"])

smtp_client.sendmail(MAIL["from"], MAIL["to"], msg.as_string())

# Tidy up
for f in ['domain_share.png','domain_share_onion.png','platform_share.png','timezones.png']:
    os.unlink(f)


'''
domain_share_ccid = make_msgid()
onion_domain_share_ccid = make_msgid()
platform_share_ccid = make_msgid()
timezone_share_ccid = make_msgid()
'''


sys.exit()
